/*
 * $Id: client.c,v 1.24 2003/03/06 14:05:34 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <errno.h>
#include <termios.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "client.h"
#include "common.h"
#include "protocol.h"
#include "buffer.h"
#include "config.h"

#ifdef NDEBUG
#define DEBUG(msg)
#else
#define DEBUG(msg) printf("%s debug: %s\n", progname, msg)
#endif

/* The environment of the process. */
extern char **environ;

/** The name of this program. */
static char *progname;

/** The command line arguments. */
static char *target;
static char *curdir;
static char *cmd;
static char **args;

/** The socket fd. */
static int sd = -1;

/** Buffers used to copy data around. */
static buffer_t buf_out;
static buffer_t buf_err;
static char tmp_buf[BUFFER_SIZE + 1];

/** The return code that will eventually be received from the server. */
static int16_t rc = 0;

/** Is the tty in raw mode? */
static int rawmode = 0;

/** The original mode of the tty. */
static struct termios oldtio;

/** How much IN DATA has the server requested? */
static int inreq = 0;

/** Have we requested some data? */
static int outwait = 0;
static int errwait = 0;

/**
 * Prints progname and message to stderr.
 */
void error(const char *msg)
{
	fprintf(stderr, "%s: %s\n", progname, msg);
}

/**
 * Prints progname, message and errno description to stderr.
 */
void error_no(const char *msg)
{
	char *desc = strerror(errno);

	if (msg) {
		fprintf(stderr, "%s: %s (%s)\n", progname, msg, desc);
	} else {
		fprintf(stderr, "%s: %s\n", progname, desc);
	}
}

/**
 * Writes (some of) buffer to a file.
 * @param wait is set to false when the buffer is empty
 * @return 0 on success, -1 on error
 */
static int write_buffer(buffer_t *buf,
			int fd,
			int *wait)
{
	/* buf is never at EOF so fd never becomes -1. */
	if (buf_write_out(buf, &fd) < 0) {
		error(fd == STDOUT_FILENO ?
		      "Can't write buffer to stdout" :
		      "Can't write buffer to stderr");
		return -1;
	}

	if (buf_is_empty(buf)) {
		*wait = 0;
	}

	return 0;
}

/**
 * Sends an IN DATA packet.
 * @return 0 usually, 1 if stdin reached EOF, -1 on error
 */
static int send_data(void)
{
	int ok = 0;
	int len;

	len = read(STDIN_FILENO, &tmp_buf, BUFFER_SIZE);
	if (len < 0) {
		error_no("Can't read from stdin");
		return -1;
	}

	if (len == 0) {
		DEBUG("Stdin hit EOF");
		ok = 1;
		inreq = 0;
	}

	if (write_packet(sd, PTYPE_IN_DATA, &tmp_buf, len) < 0) {
		error_no("Can't write IN DATA packet to socket");
		return -1;
	}

	/* At EOF: 0-0=0 */
	inreq -= len;

	return ok;
}

/**
 * Reads bytes from socket to a buffer.
 * @param len the maximum bytes to be read
 * @param buf the target buffer
 * @return 0 on success, -1 on error
 */
static int receive_stream(int len,
			  buffer_t *buf)
{
	if (len == 0) {
		error(buf == &buf_out ?
		      "Received empty OUT DATA packet (EOF)" :
		      "Received empty ERR DATA packet (EOF)");
		return -1;
	}

	if (buf_read_in(buf, sd, len) < 0) {
		error_no(buf == &buf_out ?
			 "Can't append OUT DATA packet to buffer" :
			 "Can't append ERR DATA packet to buffer");
		return -1;
	}

	return 0;
}

/**
 * Reads message from socket and prints it.
 * @param len the length of the message
 * @param type "error" or "warning"
 * @param code the return code (-1 for error, 0 for warning)
 * @return code on success, -1 on error
 */
static int receive_message(int len,
			   char *type,
			   int code)
{
	memset(&tmp_buf, '\0', sizeof (tmp_buf));

	if (read_buf(sd, &tmp_buf, len) < 0) {
		error_no("Can't read message packet");
		return -1;
	}

	fprintf(stderr, "%s remote %s: %s\n", progname, type, tmp_buf);

	return code;
}

/**
 * Reads a packet from the socket (sd) and does something about it.
 * The value of an RC packet is stored in the global rc variable.
 * @return 1 on RC, 0 on other valid packet type, -1 on error
 */
static int receive_packet(void)
{
	phead_t head;

	if (read_phead(sd, &head) < 0) {
		error("Can't read packet header from socket");
		return -1;
	}

	switch (head.type) {
	case PTYPE_IN_REQ:
		if (head.size) {
			error("IN REQ packet size is non-zero");
			return -1;
		}
		inreq = BUFFER_SIZE;
		return 0;

	case PTYPE_OUT_DATA:
		return receive_stream(head.size, &buf_out);

	case PTYPE_ERR_DATA:
		return receive_stream(head.size, &buf_err);

	case PTYPE_RC:
		rc = read_int16(sd);
		if (rc < 0) {
			error("Can't read RC packet from socket");
			return -1;
		}
		return 1;

	case PTYPE_ERROR:
		return receive_message(head.size, "error", -1);

	case PTYPE_WARN:
		return receive_message(head.size, "warning", 0);
	}

	error("Received packet has invalid type");
	return -1;
}

/**
 * Sends a request for more data.
 * @param ptype PTYPE_OUT_REQ or PTYPE_ERR_REQ
 * @param wait is set to true
 * @return 0 on success, -1 on error
 */
static int send_request(int ptype,
			int *wait)
{
	if (write_packet(sd, ptype, NULL, 0) < 0) {
		error("Can't write packet to socket");
		return -1;
	}

	*wait = 1;

	return 0;
}

/**
 * Read an AUTH OK or ERROR packet from socket.
 * @return authentication reply or -1 on error
 */
static int get_auth_reply(void)
{
	phead_t head;
	int16_t ok;

	if (read_phead(sd, &head) < 0) {
		error("Can't read packet header from socket");
		return -1;
	}

	switch (head.type) {
	case PTYPE_AUTH_OK:
		ok = read_int16(sd);
		if (ok < 0) {
			error("Can't read AUTH OK packet from socket");
			return -1;
		}
		return ok;

	case PTYPE_ERROR:
		return receive_message(head.size, "error", -1);
	}

	error("Received packet has invalid type");
	return -1;
}

/**
 * Manages stdin/stdout/stderr and waits for the return code.
 * @return 0 on success, -1 on error
 */
static int manage(void)
{
	fd_set readfds, writefds;
	int inopen = 1;
	int ok;

	if (!buf_init(&buf_out) ||
	    !buf_init(&buf_err)) {
		error_no("Can't initialize buffers");
		return -1;
	}

	while (1) {
		FD_ZERO(&readfds);
		FD_SET(sd, &readfds);
		if (inopen && inreq) {
			FD_SET(STDIN_FILENO, &readfds);
		}

		FD_ZERO(&writefds);
		if (!buf_is_empty(&buf_out) || !outwait) {
			FD_SET(STDOUT_FILENO, &writefds);
		}
		if (!buf_is_empty(&buf_err) || !errwait) {
			FD_SET(STDERR_FILENO, &writefds);
		}

		if (select(sd + 1, &readfds, &writefds, NULL, NULL) <= 0) {
			error_no("Can't select");
			return -1;
		}

		/* Read packet from socket. */
		if (FD_ISSET(sd, &readfds)) {
			ok = receive_packet();
			if (ok < 0) {
				return -1;
			}
			if (ok > 0) {
				break;
			}
		}

		/* Send data from stdin if requested. */
		if (inopen && inreq && FD_ISSET(STDIN_FILENO, &readfds)) {
			ok = send_data();
			if (ok < 0) {
				return -1;
			}
			if (ok > 0) {
				inopen = 0;
			}
		}

		/* Flush buf_out to stdout or request more data. */
		if (FD_ISSET(STDOUT_FILENO, &writefds)) {
			if (buf_is_empty(&buf_out)) {
				if (!outwait &&
				    send_request(PTYPE_OUT_REQ,&outwait) < 0) {
					return -1;
				}
			} else {
				if (write_buffer(&buf_out, STDOUT_FILENO,
						 &outwait)<0) {
					return -1;
				}
			}
		}

		/* Flush buf_err to stderr or request more data. */
		if (FD_ISSET(STDERR_FILENO, &writefds)) {
			if (buf_is_empty(&buf_err)) {
				if (!errwait &&
				    send_request(PTYPE_ERR_REQ,&errwait) < 0) {
					return -1;
				}
			} else {
				if (write_buffer(&buf_err, STDERR_FILENO,
						 &errwait)<0) {
					return -1;
				}
			}
		}
	}

	/* See that stdout and stderr buffers are flushed. */
	while (!buf_is_empty(&buf_out) &&
	       write_buffer(&buf_out, STDOUT_FILENO, &outwait) == 0);
	while (!buf_is_empty(&buf_err) &&
	       write_buffer(&buf_err, STDERR_FILENO, &errwait) == 0);

	return 0;
}

/**
 * Sets stdin to raw mode.
 */
static int set_raw_mode(void)
{
	struct termios tio;

	if (tcgetattr(STDIN_FILENO, &tio) < 0) {
		error_no("Can't get termios");
		return -1;
	}
	oldtio = tio;

	/* OpenSSH does something like this and it seems to almost work... */
	tio.c_iflag |= IGNPAR;
	tio.c_iflag &= ~(ISTRIP | INLCR | IGNCR | ICRNL | IXON | IXANY);
	tio.c_lflag &= ~(IXOFF | ISIG | ICANON | ECHO | ECHOE | ECHOK);
	tio.c_lflag &= ~(ECHONL | IEXTEN | OPOST);
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 0;

	DEBUG("Entering raw mode");

	if (tcsetattr(STDIN_FILENO, TCSADRAIN, &tio) < 0) {
		error_no("Can't change termios");
		return -1;
	}

	rawmode = 1;
	return 0;
}

/**
 * Sets stdin in its original mode.
 */
static void set_old_mode(void)
{
	if (!rawmode) {
		DEBUG("Not in raw mode");
		return;
	}

	DEBUG("Leaving raw mode");

	if (tcsetattr(STDIN_FILENO, TCSADRAIN, &oldtio) < 0) {
		error_no("Can't restore original termios");
	}
}

static void print_usage(void)
{
	fprintf(stderr, "Usage: %s target currentdir command [args]\n"
		"       %s --version|-v\n", progname, progname);
}

/**
 * Fills in the options from argv. Prints usage and exits on error.
 */
static void read_args(int argc,
		      char **argv)
{
	int i;

	for (i = 1; i < argc && i <= 3; i++) {
		if (strcmp(argv[i], "--version") == 0 ||
		    strcmp(argv[i], "-v") == 0) {
			fprintf(stderr, CLIENT_VERSION "\n");
			exit(0);
		}

		if (strcmp(argv[i], "--help") == 0 ||
		    strcmp(argv[i], "-h") == 0) {
			print_usage();
			exit(0);
		}
	}

	if (argc < 4) {
		print_usage();
		exit(1);
	}

	target = argv[1];
	curdir = argv[2];
	cmd = argv[3];
	args = &argv[3];  /* Include the command name! */
}

/**
 * Fills in the port and password from $HOME/.sbrsh
 */
static void open_config(config_t *cfg)
{
	char *home, *path;

	home = getenv("HOME");
	path = malloc(strlen(home) + strlen("/" CONFIG_NAME) + 1);
	strcpy(path, home);
	strcat(path, "/" CONFIG_NAME);

	if (!read_config(path, target, cfg)) {
		fprintf(stderr, "%s: Error while reading config file %s\n",
			progname, path);
		exit(1);
	}

	free(path);

	if (!cfg->user) {
		cfg->user = getenv("USER");
	}
	if (!cfg->port) {
		cfg->port = DEFAULT_PORT;
	}
}

/**
 * Called when the process exits.
 */
static void cleanup(void)
{
	DEBUG("Cleanup");

	if (sd >= 0) {
		close(sd);
	}

	/* This is quite important. */
	set_old_mode();
}

/**
 * Reads options, resolves and connects to host, talks with it, returns rc.
 */
int main(int argc,
	 char **argv)
{
	config_t cfg;
	struct hostent *host;
	struct sockaddr_in addr;
	int16_t is_tty, authok;

	init_config(&cfg);
	progname = get_progname(argv[0]);
	read_args(argc, argv);
	open_config(&cfg);

	/* We really want to exit nicely if we touch the tty mode later on. */
	atexit(cleanup);
	signal(SIGINT, exit);
	signal(SIGHUP, exit);
	signal(SIGTERM, exit);

	/* Should terminal emulation be supported? */
	is_tty = isatty(STDIN_FILENO) &&
		 isatty(STDOUT_FILENO) &&
		 isatty(STDERR_FILENO);
	DEBUG(is_tty ? "In tty mode" : "Not in tty mode");

	/* Lookup host. */
	host = gethostbyname(cfg.host);
	if (!host) {
		error("Can't resolve host");
		return 1;
	}

	/* Create socket. */
	sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		error_no("Can't create socket");
		return 1;
	}

	/* Connect to server. */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = *(unsigned int *) host->h_addr;
	addr.sin_port = htons(cfg.port);

	if (connect(sd, (struct sockaddr *) &addr, sizeof (addr)) < 0) {
		error_no("Can't connect");
		return 1;
	}

	/* Send authentication info to server. */
	write_auth(sd, cfg.user, cfg.pwd);

	/* Wait for a reply... */
	authok = get_auth_reply();
	if (authok < 0) {
		error("Can't read authentication reply");
		return 1;
	}

	if (!authok) {
		error("Authentication failed");
		return 1;
	}

	/* Send command info to server. */
	write_cmd(sd, target, (const char **) cfg.opts, cmd,
		  (const char **) args, curdir, (const char **) environ,
		  is_tty);

	/* We don't need this anymore. */
	free_config(&cfg);

	if (is_tty) {
		if (set_raw_mode() < 0) {
			return 1;
		}
	} else {
		if (set_nonblocking(STDIN_FILENO) < 0 ||
		    set_nonblocking(STDOUT_FILENO) < 0 ||
		    set_nonblocking(STDERR_FILENO) < 0) {
			error_no("Can't make stdin/out/err non-blocking");
			return 1;
		}
	}

	/* Manage in/out/err and wait for the rc. */
	if (manage() < 0) {
		return 1;
	}

	DEBUG(rc ? "Error return" : "Normal return");
	return rc;
}
