/*
 * $Id: mount.c,v 1.2 2003/03/07 12:06:07 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <wait.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "mount.h"
#include "common.h"
#include "daemon.h"

#define MKDIR_PERMS 0755

/**
 * Recursively creates all needed directories.
 */
int mkdirs(const char *path)
{
	char *parent, *p;
	int rc = -1;

	if (mkdir(path, MKDIR_PERMS) == 0 || errno == EEXIST) {
		return 0;
	}

	if (errno != ENOENT) {
		return -1;
	}

	parent = strdup(path);
	if (!parent) {
		return -1;
	}

	p = strrchr(parent, '/');
	if (p && parent != p) {
		*p = '\0';
		rc = 0;
	}

	if (rc == 0) {
		rc = mkdirs(parent);
	}

	free(parent);

	if (rc == 0) {
		rc = mkdir(path, MKDIR_PERMS);
	}

	return rc;
}

/**
 * Parses a string of form: <type> <device> <point> [<options>]
 */
mount_t *mnt_parse(char *str,
		   const char *root)
{
	char *pnt, *dev, *opt;
	int type;
	mount_t *m;

	/* 2nd: device */
	dev = find_space(str);
	if (*dev == '\0') {
		send_error("No device in mount string");
		return NULL;
	}
	*dev++ = '\0';
	dev = skip_spaces(dev);

	/* 3rd: point */
	pnt = find_space(dev);
	if (*pnt == '\0') {
		send_error("No mount point in mount string");
		return NULL;
	}
	*pnt++ = '\0';
	pnt = skip_spaces(pnt);

	/* 4th: options */
	opt = find_space(pnt);
	if (*opt != '\0') {
		*opt++ = '\0';
		opt = skip_spaces(opt);
	}
	if (*opt == '\0') {
		opt = NULL;
	}

	/* 1st: type */
	if (strcmp(str, "nfs") == 0) {
		type = MTYPE_NFS;
	} else if (strcmp(str, "bind") == 0) {
		type = MTYPE_BIND;
	} else {
		send_error("Invalid mount type");
		return NULL;
	}

	m = malloc(sizeof (mount_t));
	if (!m) {
		send_error("Out of memory");
		return NULL;
	}

	m->type = type;

	m->device = strdup(dev);
	if (!m->device) {
		send_error("Out of memory");
		free(m);
		return NULL;
	}

	m->point = malloc(strlen(root) + strlen(pnt) + 1);
	if (!m->point) {
		send_error("Out of memory");
		free(m->device);
		free(m);
		return NULL;
	}
	strcpy(m->point, root);
	strcat(m->point, pnt);

	/* Remove / from the end. */
	str = strrchr(m->point, '/');
	if (str && strlen(str) == 1) {
		*str = '\0';
	}

	if (opt) {
		m->opts = strdup(opt);
		if (!m->opts) {
			send_error("Out of memory");
			free(m->point);
			free(m->device);
			free(m);
			return NULL;
		}
	} else {
		m->opts = NULL;
	}

	return m;
}

void mnt_free(mount_t *m)
{
	free(m->device);
	free(m->point);
	free(m);
}

/**
 * Checks /etc/mtab if m->point is already mounted.
 */
int mnt_is_mounted(const mount_t *m)
{
	FILE *file;
	char buf[1024], *pnt;
	int len, yes = 0;

	file = fopen(MOUNT_TAB, "r");
	if (!file) {
		send_error("Can't open file " MOUNT_TAB);
		return -1;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}

		pnt = skip_spaces(find_space(buf));
		*find_space(pnt) = '\0';

		if (strcmp(pnt, m->point) == 0) {
			yes = 1;
			break;
		}
	}

	fclose(file);

	return yes;
}

/**
 * Calls /bin/mount.
 */
int mnt_mount(const mount_t *m)
{
	char *cmd;
	int len, status;

	switch (m->type) {
	case MTYPE_NFS:
		len = strlen(MOUNT_CMD) + 1 + strlen(m->device) + 1 +
		      strlen(m->point) + strlen(" -t nfs") + 1;
		if (m->opts) {
			len += strlen(" -o ") + strlen(m->opts);
		}

		cmd = malloc(len);
		strcpy(cmd, MOUNT_CMD);
		strcat(cmd, " ");
		strcat(cmd, m->device);
		strcat(cmd, " ");
		strcat(cmd, m->point);
		strcat(cmd, " -t nfs");
		if (m->opts) {
			strcat(cmd, " -o ");
			strcat(cmd, m->opts);
		}

		DEBUG("Executing: %s", cmd);

		status = system(cmd);
		if (status < 0) {
			return -1;
		}
		if (WEXITSTATUS(status) != 0) {
			errno = 0;
			return -1;
		}
		return 0;

	case MTYPE_BIND:
		len = strlen(MOUNT_CMD) + strlen(" --bind ") +
		      strlen(m->device) + 1 + strlen(m->point) + 1;

		cmd = malloc(len);
		strcpy(cmd, MOUNT_CMD);
		strcat(cmd, " --bind ");
		strcat(cmd, m->device);
		strcat(cmd, " ");
		strcat(cmd, m->point);

		DEBUG("Executing: %s", cmd);

		status = system(cmd);
		if (status < 0) {
			return -1;
		}
		if (WEXITSTATUS(status) != 0) {
			errno = 0;
			return -1;
		}
		return 0;
	}

	errno = EINVAL;
	return -1;
}

int mnt_unmount(const mount_t *m)
{
	return umount(m->point);
}
